/*
 *  MPEG-1 Real Time Encoder
 *
 *  Copyright (C) 1999-2000 Michael H. Schimek
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "ieee.h"

static double aan_fwd_lut[8][8];
static double aan_inv_lut[8][8];

static void init_lut(void) __attribute__ ((constructor));

static void
init_lut(void)
{
	int v, u;

	for (v = 0; v < 8; v++) {
		for (u = 0; u < 8; u++) {
			double Cu, Cv;

    			Cu = (u == 0) ? 1.0 : (cos(u * IEEE_PI / 16.0) * sqrt(2.0));
			Cv = (v == 0) ? 1.0 : (cos(v * IEEE_PI / 16.0) * sqrt(2.0));

			aan_fwd_lut[v][u] = 1.0 / (Cu * Cv * 8.0);
			aan_inv_lut[v][u] = 1.0 * (Cu * Cv / 8.0);
		}
	}

	dump(aan_fwd_lut);
	dump(aan_inv_lut);
}

static void
aan_double_1d_fdct(double *in, double *out)
{
	double tmp0, tmp1, tmp2, tmp3;
	double tmp4, tmp5, tmp6, tmp7;

	/* even */

    	tmp0 = in[0] + in[7];
	tmp1 = in[1] + in[6];
	tmp2 = in[2] + in[5];
	tmp3 = in[3] + in[4];

	tmp4 = tmp0 + tmp3;
        tmp5 = tmp1 + tmp2;

	out[0] = tmp4 + tmp5;
        out[4] = tmp4 - tmp5;

	tmp0 -= tmp3;
	tmp1 -= tmp2;

	tmp1 = (tmp0 + tmp1) * C4;

	out[2] = tmp0 + tmp1;
	out[6] = tmp0 - tmp1;

	/* odd */

	tmp7 = in[0] - in[7];
	tmp6 = in[1] - in[6];
	tmp5 = in[2] - in[5];
	tmp4 = in[3] - in[4];

	tmp4 += tmp5;
	tmp5 += tmp6;
        tmp6 += tmp7;

	tmp3 = (tmp4 - tmp6) * C6;
	tmp4 = (tmp4 * (C2 - C6)) + tmp3;
	tmp6 = (tmp6 * (C2 + C6)) + tmp3;
	tmp5 = tmp5 * C4;

	tmp3 = tmp7 + tmp5;
	tmp7 = tmp7 - tmp5;

	out[5] = tmp7 + tmp4;
        out[3] = tmp7 - tmp4;
        out[1] = tmp3 + tmp6;
        out[7] = tmp3 - tmp6;
}

void
aan_double_fdct(short block[8][8])
{
    double t1[8][8], t2[8][8];
    int v, u;

    for (v = 0; v < 64; v++)
	t1[0][v] = block[0][v]; 

    for (v = 0; v < 8; v++)
	aan_double_1d_fdct(t1[v], t2[v]);

    mirror(t2);

    for (u = 0; u < 8; u++)
	aan_double_1d_fdct(t2[u], t1[u]);

    mirror(t1);

    for (v = 0; v < 8; v++)
        for (u = 0; u < 8; u++)
	    block[v][u] = ieee_round(t1[v][u] * aan_fwd_lut[v][u]);
}

static void
aan_double_1d_idct(double *in, double *out)
{
	double tmp0, tmp1, tmp2, tmp3;
	double tmp4, tmp5, tmp6, tmp7, tmp8;

	/* odd */

	tmp5 = in[5] + in[3];
	tmp6 = in[5] - in[3];
	tmp8 = in[1] + in[7];
	tmp4 = in[1] - in[7];

	tmp7 = tmp8 + tmp5;
	tmp8 = tmp8 - tmp5;

	tmp5 = tmp4 + tmp6;
	tmp5 = +2.0 * C2 * tmp5;
	tmp6 = -2.0 * (C2 + C6) * tmp6 + tmp5;
	tmp4 = -2.0 * (C2 - C6) * tmp4 + tmp5;
	tmp5 = +2.0 * C4 * tmp8;

	tmp6 -= tmp7;
	tmp5 -= tmp6;
	tmp4 -= tmp5;

	/* even */

	tmp2 = in[2] + in[6];
	tmp8 = in[2] - in[6];

	tmp8 = tmp8 * 2.0 * C4 - tmp2;

	tmp0 = in[0] + in[4];
	tmp1 = in[0] - in[4];

	tmp3 = tmp0 - tmp2;
	tmp0 = tmp0 + tmp2;
	tmp2 = tmp1 - tmp8;
	tmp1 = tmp1 + tmp8;

	out[0] = tmp0 + tmp7;
	out[1] = tmp1 + tmp6;
	out[2] = tmp2 + tmp5;
	out[3] = tmp3 + tmp4;
	out[4] = tmp3 - tmp4;
	out[5] = tmp2 - tmp5;
	out[6] = tmp1 - tmp6;
	out[7] = tmp0 - tmp7;
}

void
aan_double_idct(short block[8][8])
{
    double t1[8][8], t2[8][8];
    int v, u;
	    
    for (v = 0; v < 8; v++)
        for (u = 0; u < 8; u++)
	    t1[v][u] = block[v][u] * aan_inv_lut[v][u];

    mirror(t1);

    for (u = 0; u < 8; u++)
	aan_double_1d_idct(t1[u], t2[u]);

    mirror(t2);

    for (v = 0; v < 8; v++)
	aan_double_1d_idct(t2[v], t1[v]);

    for (v = 0; v < 64; v++)
	block[0][v] = ieee_round(t1[0][v]);
}
